@@ -307,6 +307,9 @@ class SpecTypeDefn(SpecType):
     def is_base_type(self, type_: BaseTypeName) -> bool:
         return self.is_base and self.name == type_
 
+    def can_process(self, builder: SpecBuilder, data: RawDict) -> bool:
+        return True
+
     @abc.abstractmethod
     def process(self, builder: SpecBuilder, data: RawDict) -> None: ...
 
@@ -356,9 +359,9 @@ class SpecTypeDefn(SpecType):
                 property_name_case = NameCase(name_case_raw)
 
             if property_name_case != NameCase.preserve:
-                assert util.is_valid_property_name(
-                    name
-                ), f"{name} is not a valid property name"
+                assert util.is_valid_property_name(name), (
+                    f"{name} is not a valid property name"
+                )
 
             data_type = data.get("type")
             builder.ensure(data_type is not None, "missing `type` entry")
@@ -532,13 +535,8 @@ class SpecTypeDefnObject(SpecTypeDefn):
         base_type: list[SpecType] = [self.base] if self.base is not None else []
         return base_type + prop_types
 
-    def get_generic(self) -> Optional[str]:
-        if len(self.parameters) > 0:
-            assert (
-                len(self.parameters) == 1
-            ), "Only single generic parameters current supported"
-            return self.parameters[0]
-        return None
+    def get_generics(self) -> list[str]:
+        return self.parameters
 
 
 class SpecTypeDefnAlias(SpecTypeDefn):
@@ -592,21 +590,21 @@ class SpecTypeDefnUnion(SpecTypeDefn):
             self.discriminator_map = {}
             for sub_type in self.types:
                 builder.push_where(sub_type.name)
-                assert isinstance(
-                    sub_type, SpecTypeDefnObject
-                ), "union-type-must-be-object"
+                assert isinstance(sub_type, SpecTypeDefnObject), (
+                    "union-type-must-be-object"
+                )
                 assert sub_type.properties is not None
                 discriminator_type = sub_type.properties.get(self.discriminator)
-                assert (
-                    discriminator_type is not None
-                ), f"missing-discriminator-field: {sub_type}"
+                assert discriminator_type is not None, (
+                    f"missing-discriminator-field: {sub_type}"
+                )
                 prop_type = unwrap_literal_type(discriminator_type.spec_type)
                 assert prop_type is not None
                 assert prop_type.is_value_to_string()
                 discriminant = str(prop_type.value)
-                assert (
-                    discriminant not in self.discriminator_map
-                ), f"duplicated-discriminant, {discriminant} in {sub_type}"
+                assert discriminant not in self.discriminator_map, (
+                    f"duplicated-discriminant, {discriminant} in {sub_type}"
+                )
                 self.discriminator_map[discriminant] = sub_type
 
                 builder.pop_where()
@@ -666,15 +664,30 @@ class SpecTypeDefnStringEnum(SpecTypeDefn):
         self.desc: str | None = None
         self.sql_type_name: Optional[str] = None
         self.emit_id_source = False
+        self.source_enums: list[SpecType] = []
+
+    def can_process(self, builder: SpecBuilder, data: dict[Any, Any]) -> bool:
+        source_enums = data.get("source_enums")
+        try:
+            for sub_type_str in source_enums or []:
+                sub_type = builder.parse_type(self.namespace, sub_type_str)
+                assert isinstance(sub_type, SpecTypeDefnStringEnum)
+                assert len(sub_type.values) > 0
+        except AssertionError:
+            return False
+        return super().can_process(builder, data)
 
     def process(self, builder: SpecBuilder, data: RawDict) -> None:
         super().base_process(
-            builder, data, ["type", "desc", "values", "name_case", "sql", "emit"]
+            builder,
+            data,
+            ["type", "desc", "values", "name_case", "sql", "emit", "source_enums"],
         )
         self.name_case = NameCase(data.get("name_case", "convert"))
         self.values = {}
-        data_values = data["values"]
+        data_values = data.get("values")
         self.desc = data.get("desc", None)
+        source_enums = data.get("source_enums", None)
         if isinstance(data_values, dict):
             for name, value in data_values.items():
                 builder.push_where(name)
@@ -712,10 +725,13 @@ class SpecTypeDefnStringEnum(SpecTypeDefn):
         elif isinstance(data_values, list):
             for value in data_values:
                 if value in self.values:
-                    raise Exception("duplicate value in typespec enum", self.name, value)
+                    raise Exception(
+                        "duplicate value in typespec enum", self.name, value
+                    )
                 self.values[value] = StringEnumEntry(name=value, value=value)
         else:
-            raise Exception("unsupported values type")
+            if source_enums is None or data_values is not None:
+                raise Exception("unsupported values type")
 
         sql_data = data.get("sql")
         if sql_data is not None:
@@ -737,9 +753,18 @@ class SpecTypeDefnStringEnum(SpecTypeDefn):
                 builder.ensure(
                     entry.label is not None, f"need-label-for-id-source:{entry.name}"
                 )
+        for sub_type_str in source_enums or []:
+            sub_type = builder.parse_type(self.namespace, sub_type_str)
+            self.source_enums.append(sub_type)
+
+        for sub_type in self.source_enums:
+            builder.push_where(sub_type.name)
+            if isinstance(sub_type, SpecTypeDefnStringEnum):
+                self.values.update(sub_type.values)
+            builder.pop_where()
 
     def get_referenced_types(self) -> list[SpecType]:
-        return []
+        return self.source_enums
 
 
 TOKEN_ENDPOINT = "$endpoint"
@@ -899,9 +924,9 @@ class SpecEndpoint:
         self.is_external = self.path_root == "api/external"
         self.has_attachment = data.get("has_attachment", False)
 
-        assert (
-            not is_sdk or self.desc is not None
-        ), f"Endpoint description required for SDK endpoints, missing: {path}"
+        assert not is_sdk or self.desc is not None, (
+            f"Endpoint description required for SDK endpoints, missing: {path}"
+        )
 
     @property
     def resolved_path(self: Self) -> str:
@@ -972,7 +997,9 @@ def _parse_const(
             return value
 
         if const_type.name == BaseTypeName.s_boolean:
-            builder.ensure(isinstance(value, bool), "invalid value for boolean constant")
+            builder.ensure(
+                isinstance(value, bool), "invalid value for boolean constant"
+            )
             return value
 
     raise Exception("unsupported-const-scalar-type", const_type)
@@ -1004,7 +1031,9 @@ class SpecConstant:
             assert isinstance(self.value, dict)
             # the parsing checks that the values are correct, so a simple length check
             # should be enough to check completeness
-            builder.ensure(len(key_type.values) == len(self.value), "incomplete-enum-map")
+            builder.ensure(
+                len(key_type.values) == len(self.value), "incomplete-enum-map"
+            )
 
 
 class SpecNamespace:
@@ -1079,9 +1108,9 @@ class SpecNamespace:
                 continue
 
             if "value" in defn:
-                assert util.is_valid_property_name(
-                    name
-                ), f"{name} is not a valid constant name"
+                assert util.is_valid_property_name(name), (
+                    f"{name} is not a valid constant name"
+                )
                 spec_constant = SpecConstant(self, name)
                 self.constants[name] = spec_constant
                 continue
@@ -1100,7 +1129,11 @@ class SpecNamespace:
                 spec_type = SpecTypeDefnStringEnum(self, name)
             else:
                 parameters = (
-                    [parameter.name for parameter in parsed_name.parameters[0]]
+                    [
+                        parameter.name
+                        for name_parameters in parsed_name.parameters
+                        for parameter in name_parameters
+                    ]
                     if parsed_name.parameters is not None
                     else None
                 )
@@ -1116,28 +1149,41 @@ class SpecNamespace:
         Complete the definition of each type.
         """
         builder.push_where(self.name)
-        for full_name, defn in data.items():
-            parsed_name = parse_type_str(full_name)[0]
-            name = parsed_name.name
+        items_to_process: list[NameDataPair] = [
+            NameDataPair(full_name=full_name, data=defn)
+            for full_name, defn in data.items()
+        ]
+        while len(items_to_process) > 0:
+            deferred_items: list[NameDataPair] = []
+            for item in items_to_process:
+                full_name = item.full_name
+                defn = item.data
+                parsed_name = parse_type_str(full_name)[0]
+                name = parsed_name.name
 
-            if name in [TOKEN_EMIT_IO_TS, TOKEN_EMIT_TYPE_INFO, TOKEN_IMPORT]:
-                continue
+                if name in [TOKEN_EMIT_IO_TS, TOKEN_EMIT_TYPE_INFO, TOKEN_IMPORT]:
+                    continue
 
-            builder.push_where(name)
+                builder.push_where(name)
 
-            if "value" in defn:
-                spec_constant = self.constants[name]
-                spec_constant.process(builder, defn)
+                if "value" in defn:
+                    spec_constant = self.constants[name]
+                    spec_constant.process(builder, defn)
 
-            elif name == TOKEN_ENDPOINT:
-                assert self.endpoint
-                self.endpoint.process(builder, defn)
+                elif name == TOKEN_ENDPOINT:
+                    assert self.endpoint
+                    self.endpoint.process(builder, defn)
 
-            else:
-                spec_type = self.types[name]
-                spec_type.process(builder, defn)
+                else:
+                    spec_type = self.types[name]
+                    if spec_type.can_process(builder, defn):
+                        spec_type.process(builder, defn)
+                    else:
+                        deferred_items.append(item)
 
-            builder.pop_where()
+                builder.pop_where()
+            assert len(deferred_items) < len(items_to_process)
+            items_to_process = [deferred for deferred in deferred_items]
 
         builder.pop_where()
 
@@ -1152,8 +1198,15 @@ class NamespaceDataPair:
     data: RawDict
 
 
+@dataclass(kw_only=True)
+class NameDataPair:
+    full_name: str
+    data: RawDict
+
+
 class SpecBuilder:
-    def __init__(self, *, api_endpoints: dict[str, str]) -> None:
+    def __init__(self, *, api_endpoints: dict[str, str], top_namespace: str) -> None:
+        self.top_namespace = top_namespace
         self.where: list[str] = []
         self.namespaces = {}
         self.pending: list[NamespaceDataPair] = []
@@ -1335,7 +1388,9 @@ class SpecBuilder:
     ) -> SpecType:
         self.push_where(spec)
         parsed_type = util.parse_type_str(spec)
-        result = self._convert_parsed_type(parsed_type, namespace, top=True, scope=scope)
+        result = self._convert_parsed_type(
+            parsed_type, namespace, top=True, scope=scope
+        )
         self.pop_where()
         return result
 
@@ -1385,7 +1440,9 @@ class SpecBuilder:
         path_meta: list[str] | None = meta.get("path")
         guide_key: SpecGuideKey = RootGuideKey()
         if path_meta is not None:
-            path_details = _resolve_endpoint_path("".join(path_meta), self.api_endpoints)
+            path_details = _resolve_endpoint_path(
+                "".join(path_meta), self.api_endpoints
+            )
             guide_key = EndpointGuideKey(path=path_details.resolved_path)
 
         self.guides[guide_key].append(
